/* 
 * tree.c
 *
 * Function definitions for the tree 
 *
 * Changes:
 *  04/05/07 - Header creation (jung@ucsd.edu)
 */
#include "tree.h"

char int2char(int tt) { return (tt); }
int postorder_align(TNODE* current, char* profile, int prof_len, 
                    char* sequence, int seq_len, int** grid, TNODE** best,
                    int best_score, int** CC) 
{
	int i, j;
	int compact_prof;     // compact representation of the profile
	int consensus;
	int unused;
	int score = MAXSCORE, rec_score = MAXSCORE, right_score;

	if (current->lchild) {
	   rec_score = postorder_align(current->lchild, profile, prof_len,
	                              sequence, seq_len, grid, best, best_score, CC);
	}
	if (current->rchild) {
	   if (best_score > rec_score)   best_score = rec_score;
	   rec_score = postorder_align(current->rchild, profile, prof_len,
	                                 sequence, seq_len, grid, best, best_score, CC);
	}

	if (current->father)
	{
	   // non root node, align this with the egde between this and parent
	   for(i=0; i<prof_len; i++) 
		{
		   compact_prof = 0;
		   for (j=4; j>=0; j--) 
			{
				if( in(j, current->profile[i]) || in(j, current->father->profile[i]))
		         consensus = 1;
		      else consensus = 0;
	   	   compact_prof = compact_prof*2 + consensus;
	   	}
			//profile of edge(current-->current.father) with profile[0]=0;
	   	profile[i+1] = int2char(compact_prof);
	   }
	   if(current->psa == 1 || QUICKHIT == 0) {
	   	//when QUICKHIT==1 and current->psa is set to 1
	   	//or when QUICKHIT==0
		   score = align_bending(profile, sequence, prof_len, seq_len, CC, grid, &unused);
		   current->score = score;
	   }
//	   printf ("TNODE=%x, score=%d\n", &current, score);
	}

	if (score < rec_score && score < best_score) {
		// we found a better match
//	   printf ("current score < best_score\n");
		*best = current;
		return score;
	} else {
//		printf ("\n");
	}

	if (rec_score < best_score) return rec_score;
	return best_score;
}



/*
 * Continue search from current until reach local optimum
 */
int continue_align (TNODE* current, char* profile, int prof_len,
                    char* sequence, int seq_len, int ** grid,
                    int best_score, int** CC) {
}


/*
 * Print the tree in newlick format. Use in order recursion.
 */
void print_newlick(FILE* fout, char** names, TNODE* current) {
  if (current->seqid >= 0) {
    fprintf(fout, "%s:%.5f", names[current->seqid], current->weight);
  }
  else {
    fprintf(fout, "(");
    print_newlick(fout, names, current->lchild);
    fprintf(fout, ",");
    print_newlick(fout, names, current->rchild);
	if(current->father !=0)
	{
		fprintf(fout, "):%.5f", current->weight);
	}
	else
	{
		fprintf(fout, ");");
	}
  }
  return;
}


/*
 * Insert a node into our tree. The insertion will take place between target 
 * and its parent. insert is the sequence to insert.
 */
void insert_node(TNODE* target, TNODE* insert, int length) {
	if (target->father) 
	{
		// create a new internal node
		TNODE* internal = (TNODE*)calloc(1, sizeof(TNODE));
		inorder_allocate(internal, length);
		internal->seqid = -1;
		internal->father = target->father;
		internal->lchild = insert;
		internal->rchild = target;

		if (target->father->lchild == target) 
		// target is a left child
		target->father->lchild = internal;
		else
		// target is a right child
		target->father->rchild = internal;

		//missing:
		insert->father = internal;
		target->father = internal;	
	}
	else 
	{
		// this is impossible because root has no father
		printf("ERROR: Attempt to insert between root and its parent!\n");
		exit(-1);
	}
}


/* print subtrees in broadth first order, current should be the root of the subtree*/
void printsubtreeBFS(FILE* fout, char** names, TNODE* current) {
	assert(current);
	queue<TNODE*> qt;
	qt.push(current);
	
	while(!qt.empty()) {
		TNODE* n = qt.front();
		qt.pop();
		
		if(n->lchild && n->rchild) {
			fprintf(fout, "[v%d]\t", n->nodeindx);
			print_newlick(fout, names, n);
			fprintf(fout, "\n");
		}
		if(n->lchild) {qt.push(n->lchild);}
		if(n->rchild) {qt.push(n->rchild);}
	}
	
}

/* print fitch score profile of each internal node, current should be the root of the subtree*/
void printfitchBFS(FILE* fout, int length, TNODE* current) {
	assert(current);
	
	queue<TNODE*> qt;
	qt.push(current);
	while(!qt.empty()) {
		TNODE* n = qt.front();
		qt.pop();
		
		int sum = 0;
		if(n->lchild && n->rchild) {
		//don't print out fitch score profile for leaf nodes.
			fprintf(fout, "[v%d]\t", n->nodeindx);
			for(int i=0; i<length; i++) {
				sum += n->cost[i];
				fprintf(fout, "%d\t", n->cost[i]);
			}
			fprintf(fout, "%d\n", sum);
		}
		if(n->lchild) qt.push(n->lchild);
		if(n->rchild) qt.push(n->rchild);
	}
	
}

/* assign internal node index in broadth first order, current should be the root of the subtree, current->nodeindx=indx*/
void assignnodeindxBFS(TNODE* current, int indx, char** tname) {
	assert(current);

	queue<TNODE*> qt;
	qt.push(current);
	
	while(!qt.empty()) {
		TNODE* n = qt.front();
		qt.pop();

		
		if(n->lchild && n->rchild) {
			n->nodeindx = indx;
			indx++;
//			printf("visiting internal node %d\n", n->nodeindx);
		}else{
			n->nodeindx = -1;//for leaves, nodeindx = -1 
//			printf("visiting leaf node %s\n", tname[n->seqid]);			
		}
		
		if(n->lchild) qt.push(n->lchild);
		if(n->rchild) qt.push(n->rchild);
	}
}

/*
 * In order allocate the space for the tree nodes
 */
void inorder_allocate(TNODE* current, int length, int allocate_profile) {
  int j;
	if(!current) return;
	current->first = (char*) malloc(length * sizeof(char));//(int**)malloc(length * sizeof(int*));
	current->second = (char*) malloc(length * sizeof(char));//(int**)malloc(length * sizeof(int*));
	if(allocate_profile) current->profile = (char*) malloc(length * sizeof(char));//current->profile = (int**)malloc(length * sizeof(int*));
	current->cost = (int*)malloc(length * sizeof(int));

	if(!current->first|| !current->second || !current->cost || (!current->profile && allocate_profile))
	{
		printf("Error!! Can not allocate memory!!\n");
	} 

	for (j=0; j< length; j++) {
	   current->first[j] = 0;
	   current->second[j] = 0;
	   if(allocate_profile) current->profile[j] = 0;
	   current->cost[j] = 0;
	}

	if (current->lchild)  inorder_allocate(current->lchild, length);
	if (current->rchild)  inorder_allocate(current->rchild, length);
}


/*
 * In order free free the space allocate by the reciprocal function (inorder_allocate)
 */
void inorder_free(TNODE* current) {
	int i;
	assert(current);
	if(current->first) free(current->first);
	if(current->second) free(current->second);
	if(current->profile) free(current->profile);
	if(current->cost) free(current->cost);
	current->first = 0;
	current->second = 0;
	current->profile = 0;
	current->cost = 0;

	if(current->lchild)  inorder_free(current->lchild);
	if (current->rchild)  inorder_free(current->rchild);
}


/*
 * Build the tree given the information in array form.
 * We do not do any error checking on tnode.
 *	The tree referenced by TNODE* root has not set brother field
 * Note: In mtree, children are stored before parent: 
 * 	  i.e., children have indexes less than parent.
 * 	  Thus, the root is stored at the last location
 */
int buildtree(TNODE** root, TLIST *mtree, int *tleaf, int *tnode, int numseq, int numnode, double* tweight) {
	int n = 1;

	//m: counting number of sequences (leaves) parsed from tnode
	int m = 0;
	int p = 0;
	int i, j, k;

	TLIST *tstack;

	LNODE* top_stack = 0; // the stack where the temporary nodes are

	int temp1, temp2;
	int node1, node2;

	tstack = (TLIST *) malloc (numnode * sizeof (TLIST));
	int index = 0;
	
	//mtree[0] is not an actual node of the tree
	mtree[0].flag = -1;
	for (i= 0; i < numnode; i++)  
	{

		if (tnode[i] == -1)   {
		// open parenthesis, do nothing
		}
		else if (tnode[i] == -2) 
		{
			LNODE* temp_lnode;          // for swapping
			TNODE* new_tnode;           // temp pointer of a tree node 

			// close parenthesis, pop 2 elements and merge them
			new_tnode = (TNODE*)calloc(1, sizeof(TNODE)); 
			new_tnode->seqid = -1;      // this is an internal node

			// pop 2 and used them as left and right child
			new_tnode->rchild = top_stack->data;
			new_tnode->lchild = top_stack->prev->data;

			// also complete their parent field
			new_tnode->rchild->father = new_tnode->lchild->father = new_tnode;

			// pop 1 from stack, free it and update the top of the stack
			temp_lnode = top_stack->prev;
			free(top_stack);
			top_stack = temp_lnode;

			// push the new node into the top of the stack
			top_stack->data = new_tnode;
			
			//mtree[n] = left child of tnode[i]
			mtree[n].seqid = tstack[index-2].seqid;
			mtree[n].weight = tweight[n];//tstack[index-2].weight;
			mtree[n].left = tstack[index-2].left;
			mtree[n].right = tstack[index-2].right;
			if (mtree[n].left >= 0)    {
			   mtree[mtree[n].left].father = n;
			}
			if (mtree[n].right >= 0) {
			   mtree[mtree[n].right].father = n;
			}
			if ((mtree[n].right == -1) && (mtree[n].left == -1)) {
			   tleaf[mtree[n].seqid] = n;
			   m++;
			}
			mtree[n].brother = n+1;
			mtree[n].flag = 0;
			mtree[n].level = -1;
			mtree[n].numleaf = tstack[index-2].numleaf;

			n++;
			//mtree[n] = right child of tnode[i]
			mtree[n].seqid = tstack[index-1].seqid;
			mtree[n].weight = tweight[n];//tstack[index-1].weight;
			mtree[n].left = tstack[index-1].left;
			mtree[n].right = tstack[index-1].right;
			if (mtree[n].left >= 0) {
			   mtree[mtree[n].left].father = n;
			}
			if (mtree[n].right >= 0) {
			   mtree[mtree[n].right].father = n;
			}
			if ((mtree[n].right == -1) && (mtree[n].left == -1)) {
			   tleaf[mtree[n].seqid] = n;
			   m++;
			}
			mtree[n].brother = n-1;
			mtree[n].flag = 0;
			mtree[n].level = -1;
			mtree[n].numleaf = tstack[index-1].numleaf;

			n++;
			index--;
			index--;
			tstack[index].seqid = -1;
			tstack[index].left = n-2;
			tstack[index].right = n-1;
			tstack[index].numleaf = 2;
			
			index++;
		}//end if(tnode[i]==-2)
		else if (tnode[i] >= 0) {
		// a valid (leaf) node, simply create a new list node and push it onto the stack
		LNODE* temp_lnode = (LNODE*)malloc(sizeof(LNODE));
		temp_lnode->data = (TNODE*)calloc(1, sizeof(TNODE));
		temp_lnode->data->seqid = tnode[i];

		// push the temp_lnode into the stack
		temp_lnode->prev = top_stack;
		top_stack = temp_lnode;

		tstack[index].seqid = tnode[i];
		tstack[index].left = -1;
		tstack[index].right = -1;
		tstack[index].numleaf = 0;
		
		index++;
		}
  }//end for i=0...numnode-1

  if (index == 1 ) 
	{
    // we are left with only one element in our stack, namely the root
    *root = top_stack->data;

    free(top_stack);
    top_stack = 0;

    mtree[n].seqid = tstack[0].seqid;
    mtree[n].weight = 0.0; //tstack[0].weight;
    mtree[n].left = tstack[0].left;
    mtree[n].right = tstack[0].right;
    if (mtree[n].left >= 0 ) {
      mtree[mtree[n].left].father = n;
    }
    if (mtree[n].right >= 0) {
      mtree[mtree[n].right].father = n;
    }
    mtree[n].brother = -1;
    mtree[n].father = -1;
    mtree[n].flag = 0;
    mtree[n].level = -1;
    mtree[n].numleaf = tstack[0].numleaf;
    n++;
  }
	if (m != numseq)        {
    printf ("Error: The number of leaves on the tree should be exact the same as the number of sequences in the MSA.\n");
    exit(-1);
  }
  free (tstack);
  return (n);
}


int removeleaf (TLIST *mtree, int numseq, int numnode, int n)
{
	int fa, bro, gfa;
	int root;

	fa = mtree[n].father;
	bro = mtree[n].brother;
	gfa = mtree[fa].father;

	root = numnode -1;
	if (fa >=0 && bro >=0 && gfa >= 0)      {
		if (mtree[gfa].left == fa )     {
			mtree[gfa].left = bro;
			mtree[bro].father = gfa;
			mtree[fa].father = -1;
			mtree[fa].flag = -1;
			mtree[fa].left = -1;
			mtree[fa].right = -1;
			mtree[n].father = -1;
			mtree[n].flag = -1;
	} else  if (mtree[gfa].right == fa)     {
			mtree[gfa].right = bro;
			mtree[bro].father = gfa;
			mtree[fa].father = -1;
			mtree[fa].flag = -1;
			mtree[fa].left = -1;
			mtree[fa].right = -1;
			mtree[n].father = -1;
			mtree[n].flag = -1;
		 }
	} else if (gfa < 0 && bro >=0)  {
			mtree[bro].father = -1;
			mtree[fa].father = -1;
			mtree[fa].left = -1;
			mtree[fa].right = -1;
			mtree[fa].flag = -1;
			mtree[n].flag = -1;
			mtree[n].father = -1;
			root = bro;
	} else if (gfa >= 0 && bro < 0 )        {
		 if (mtree[gfa].left == fa )     {
         mtree[gfa].left = -1;
		 } else if (mtree[gfa].right == fa)     {
         mtree[gfa].right = -1;
		 }
		 mtree[fa].father = -1;
		 mtree[fa].left = -1;
		 mtree[fa].right = -1;
		 mtree[n].father = -1;
		 mtree[fa].flag = -1;
		 mtree[n].flag = -1;
	}
	return (root);
}

/* search leaf by seqid, for all nodes, seqid>0 iff the node is a leaf node*/
TNODE* searchleafbyseqid(TNODE* current, const int seqid){
	assert(current);
	if(!current->lchild && !current->rchild)
		if(current->seqid == seqid)
			return current;
		else
			return NULL;
	TNODE* ret = NULL;
	if(current->lchild)
		ret = searchleafbyseqid(current->lchild, seqid);
	if(ret != NULL)
		return ret;
	if(current->rchild)
		ret = searchleafbyseqid(current->rchild, seqid);
	return ret;
}

/* search leaf by leaf name return TNODE* */
TNODE* searchleafbyname(TNODE* current, const char* queryname, char** names){
	assert(current);
	if(!current->lchild && !current->rchild) 
		if(strcmp(names[current->seqid],queryname)==1)
			return current;
		else return NULL;
	TNODE* ret = NULL;
	if(current->lchild)
		ret = searchleafbyname(current->lchild, queryname, names);
	if(ret != NULL)
		return ret;
	if(current->rchild)
		ret =  searchleafbyname(current->rchild, queryname, names);
	return ret;
}


/*get all the leaf node and store them in leafvec*/
void getLeafVec(TNODE* curNode, vector<TNODE*> &leafvec)
{
	assert(curNode);
	if(!curNode->lchild && !curNode->rchild)//curNode is a leaf node
		leafvec.PB(curNode);
	if(curNode->lchild)	getLeafVec(curNode->lchild, leafvec);
	if(curNode->rchild)	getLeafVec(curNode->rchild, leafvec);
}

/* get  the sequence distance between char * s1 and char* s2 */
double caldist(char* s1, char* s2, int length) {
	//each character in sequence represents a set of nucleotides(GATC-)
	//if any nucleotides in i-th character of s1 is also in i-th character of s2, the distance is 0 ; 
	//otherwise distance is 1;
	int i; 
	double distance = 0;
	for (i=0; i<length; i++) {
		if ((s1[i]&s2[i])==0) {}//distance is 0
		else {
			distance += 1; 
		}
	}
	return distance;
}	


void assignweight(TNODE* root, double * tweight, int & index) {
	if(root->lchild && root->rchild) {
		assignweight(root->lchild, tweight, index);
		assignweight(root->rchild,tweight, index);
	}
	root->weight = tweight[index];
	index++;
}


/* for each node on the tree, compute the total number of descendent leaf nodes */
void caltotalnumleaf(TNODE* current) {
	assert(current);
	
	current->psa = 0; //initialize psa along the way
	current->score = oo;
	
	current->totalnumleaf = 0;
	if(!current->lchild && !current->rchild)
		current->totalnumleaf = 1;
		
	if(current->lchild) {
		caltotalnumleaf (current->lchild);
		current->totalnumleaf += current->lchild->totalnumleaf;
	}
	if(current->rchild) {
		caltotalnumleaf (current->rchild);
		current->totalnumleaf += current->rchild->totalnumleaf;
	}
//	printf( "%p, %d\n", &current, current->totalnumleaf);
}


/* 
void decide(TNODE* current, int num){
	assert(current);
	TNODE* lc = current->lchild;
	TNODE* rc = current->rchild;
	TNODE* fa = current->father;
}
*/

/* for each node on the subtree rooted at current, set psa to be val*/
void setpsa(TNODE* current, int val) {
	assert(current);
	
	current->psa = val;
	if(current->lchild)
		setpsa(current->lchild,val);
	if(current->rchild)
		setpsa(current->rchild,val);
}



/*for each node, compute its distance to a given node (save in dist2givennode)
  and set psa if the distance is less than searchradius, return the number of psa=1
*/
int cmpdist_setpsa(TNODE* current, int direction) {
	//0 - starting point
	//1 - traverse from lchild to current, go out to rchild and father
	//2 - traverse from rchild to current, go out to lchild and father
	//3 - traverse from father to current, go out to lchild and rchild
	int ret = 0;
	assert(current);
	if(direction == 0) {
		current->dist2givennode = 0.0;
		if(current->rchild) {
			printf("Error, the starting point should be a leaf node\n");
			exit(0);
		}
	}
	TNODE * fa = current->father;
	TNODE * lc = current->lchild;
	TNODE * rc = current->rchild;
	
	
	if(direction == 1) 
		current->dist2givennode = lc->dist2givennode+lc->weight;
	else if(direction == 2)
		current->dist2givennode = rc->dist2givennode+rc->weight;
	else if(direction == 3) 
		current->dist2givennode = fa->dist2givennode+current->weight;
		
	if(current->dist2givennode <= searchradius) 
		current->psa = 1, ret += 1;
	
	if(!fa) {
		//no father, come to the root, stop	
	} else {//father exists
		if(direction == 1 || direction == 2 || direction == 0) {
			//before visiting fa, figure out current node is lchild or rchild of fa
			if(fa->lchild == current) {
				ret += cmpdist_setpsa(fa, 1);
			} else if(fa->rchild == current) {
				ret += cmpdist_setpsa(fa, 2);
			} else {
				printf("Strange, why current node is neither lchild nor rchild of its father\n");
				exit(0);
			}
		}
	}
	
	if(!lc) {
		//no lchild, stop;
	} else{
		if(direction == 2 || direction == 3) {
			ret += cmpdist_setpsa(lc, 3);
		}
	}
	
	if(!rc) {
		//no rchild, stop;
	} else {
		if(direction == 1 || direction == 3) {
			ret += cmpdist_setpsa(rc, 3);
		}
	}
	return ret;
}














































































